from enum import IntEnum


class PackageId(IntEnum):
    DATASTREAM = 1
    BATTERY_STATUS = 2
    DEVICE_INFO = 3
    BUTTON_EVENT = 4
    DEVICE_MODE = 5
    IDENTIFY = 6
    RECENTER = 7
    DISPLAY_FRAME = 8
    RAW_DATA = 9


class PackageType(IntEnum):
    REQUEST = 1
    RESPONSE = 2
    STREAM = 3


class DeviceMode(IntEnum):
    PRESET = 100
    SOFTWAVE = 101
    WAVEFRONT = 102
    API = 103


class ButtonId(IntEnum):
    """Which button is a package representing"""

    TOP = 0
    MIDDLE = 1
    BOTTOM = 2


class ButtonAction(IntEnum):
    """The action of a button package

    Up: button was released
    Down: button was pressed down
    Long: a long press was detected (button down for 500ms or more)
    LongUp: a long press was released (I think you'll always receive an "up" event with this one)
    ExtraLong: an extra long press was detected (I think it's about 2000ms, you'll always receive a "long" event
               before this one)
    ExtraLongUp: button released after an "extra long" press
    Click: button was pressed down and released "quickly" (i.e. not triggering a long press)
    DoubleClick: not used
    """

    UP = 0
    DOWN = 1
    LONG = 2
    LONGUP = 3
    EXTRALONG = 4
    EXTRALONGUP = 5
    CLICK = 6
    DOUBLECLICK = 7



from .organization import DataPackage, RawDataPackage, ButtonEvent  # noqa: F401
from .enums import ButtonAction, ButtonId  # noqa: F401
from .points import Euler3d, Point3d, Quaternion  # noqa: F401



import struct
from dataclasses import Field, dataclass, field
from struct import unpack_from
from typing import Optional, Union

from genki_wave.data.enums import ButtonAction, ButtonId, PackageId, PackageType
from genki_wave.data.points import Euler3d, Point3d, Quaternion, rotate_vector


@dataclass(frozen=True)
class PackageMetadata:
    """
    This is called `Query` in the public types header file. Splits out and holds the metadata from each package received
    """

    _fmt = "<BBH"

    type: PackageType
    id: PackageId
    payload_size: int

    @classmethod
    def from_raw_bytes(cls, raw_bytes: Union[bytearray, bytes]) -> "PackageMetadata":
        # TODO(robert): Check annotation `bytes`
        q_type, q_id, q_payload_size = unpack_from(cls._fmt, raw_bytes, 0)
        return cls(type=q_type, id=q_id, payload_size=q_payload_size)

    @classmethod
    def split_out_data_from_metadata(cls, raw_bytes: Union[bytearray, bytes]) -> Union[bytearray, bytes]:
        return raw_bytes[struct.calcsize(cls._fmt) :]

    def to_bytes(self) -> bytes:
        return struct.pack(self._fmt, self.type, self.id, self.payload_size)

    def is_data(self):
        return self.id == PackageId.DATASTREAM

    def is_button(self):
        return self.id == PackageId.BUTTON_EVENT

    def is_raw_gyro_accel(self):
        return self.id == PackageId.RAW_DATA


@dataclass(frozen=True)
class DataPackage:
    """Represents a data package sent from wave

    Note: Initializing DataClasses is (relatively) slow, if in the unlikely event this becomes a bottleneck,
          refactor into a new structure
    """

    _raw_len = 105
    _raw_gyro_accel_len = 32

    gyro: Point3d
    acc: Point3d
    mag: Point3d
    raw_pose: Quaternion
    current_pose: Quaternion
    euler: Euler3d
    linacc: Point3d
    peak: bool
    peak_norm_velocity: float
    timestamp_us: int
    grav: Point3d = field(init=False)
    acc_glob: Point3d = field(init=False)
    linacc_glob: Point3d = field(init=False)

    def __post_init__(self):
        # A way to initialize a derived field in a frozen dataclass
        super().__setattr__("grav", self.acc - self.linacc)
        super().__setattr__("acc_glob", rotate_vector(self.acc, self.current_pose))
        super().__setattr__("linacc_glob", rotate_vector(self.linacc, self.current_pose))

    @classmethod
    def from_raw_bytes(cls, data: Union[bytearray, bytes]) -> "DataPackage":
        # Explanation for the byte structure: https://docs.python.org/3/library/struct.html
        assert len(data) == cls._raw_len, f"Expected the raw data to have len={cls._raw_len}, got len={len(data)}"
        # These parameters encode how to read the bytes from the stream
        return cls(
            gyro=Point3d(*unpack_from("<3f", data, 0)),
            acc=Point3d(*unpack_from("<3f", data, 12)),
            mag=Point3d(*unpack_from("<3f", data, 24)),
            raw_pose=Quaternion(*unpack_from("<4f", data, 36)),
            current_pose=Quaternion(*unpack_from("<4f", data, 52)),
            euler=Euler3d(*unpack_from("<3f", data, 68)),
            linacc=Point3d(*unpack_from("<3f", data, 80)),
            peak=unpack_from("?", data, 92)[0],
            peak_norm_velocity=unpack_from("<f", data, 93)[0],
            timestamp_us=unpack_from("<Q", data, 97)[0],
        )

    def as_dict(self) -> dict:
        # This is (and should be) equivalent to `asdict(self)`, but is about 20-30x faster since it doesn't have
        # to recursively expand all dataclass fields
        return {
            "gyro": self.gyro.as_dict(),
            "acc": self.acc.as_dict(),
            "mag": self.mag.as_dict(),
            "raw_pose": self.raw_pose.as_dict(),
            "current_pose": self.current_pose.as_dict(),
            "euler": self.euler.as_dict(),
            "linacc": self.linacc.as_dict(),
            "peak": self.peak,
            "peak_norm_velocity": self.peak_norm_velocity,
            "timestamp_us": self.timestamp_us,
            "grav": self.grav.as_dict(),
            "acc_glob": self.acc_glob.as_dict(),
            "linacc_glob": self.linacc_glob.as_dict(),
        }

    def as_flat_dict(self) -> dict:
        # This is (and should be) equivalent to `flatten_nested_dicts(asdict(self))`, but is about 20-30x faster since
        # it doesn't have to recursively expand all dataclass fields
        return {
            **self.gyro.as_dict("gyro_"),
            **self.acc.as_dict("acc_"),
            **self.mag.as_dict("mag_"),
            **self.raw_pose.as_dict("raw_pose_"),
            **self.current_pose.as_dict("current_pose_"),
            **self.euler.as_dict("euler_"),
            **self.linacc.as_dict("linacc_"),
            "peak": self.peak,
            "peak_norm_velocity": self.peak_norm_velocity,
            "timestamp_us": self.timestamp_us,
            **self.grav.as_dict("grav_"),
            **self.acc_glob.as_dict("acc_glob_"),
            **self.linacc_glob.as_dict("linacc_glob_"),
        }

    @classmethod
    def flat_keys(cls) -> tuple:
        return tuple(flatten_nested_dataclass_fields(cls, None))


@dataclass(frozen=True)
class RawDataPackage:
    """Represents a package of raw data (just acc, gyro, and timestamp) sent from wave"""

    _raw_len = 32

    gyro: Point3d
    acc: Point3d
    timestamp_us: int

    @classmethod
    def from_raw_bytes(cls, data: Union[bytearray, bytes]) -> "RawDataPackage":
        # Explanation for the byte structure: https://docs.python.org/3/library/struct.html
        if len(data) != cls._raw_len:
            raise ValueError(f"Expected the raw gyro/accel data to have len={cls._raw_len}, got len={len(data)}", data)

        # These parameters encode how to read the bytes from the stream
        return cls(
            gyro=Point3d(*unpack_from("<3f", data, 0)),
            acc=Point3d(*unpack_from("<3f", data, 12)),
            timestamp_us=unpack_from("<Q", data, 24)[0],
        )

    def as_dict(self) -> dict:
        # This is (and should be) equivalent to `asdict(self)`, but is about 20-30x faster since it doesn't have
        # to recursively expand all dataclass fields
        return {"gyro": self.gyro.as_dict(), "acc": self.acc.as_dict(), "timestamp_us": self.timestamp_us}

    def as_flat_dict(self) -> dict:
        # This is (and should be) equivalent to `flatten_nested_dicts(asdict(self))`, but is about 20-30x faster since
        # it doesn't have to recursively expand all dataclass fields
        return {**self.gyro.as_dict("gyro_"), **self.acc.as_dict("acc_"), "timestamp_us": self.timestamp_us}

    @classmethod
    def flat_keys(cls) -> tuple:
        return tuple(flatten_nested_dataclass_fields(cls, None))


@dataclass(frozen=True)
class ButtonEvent:
    """Represents a button event sent from wave"""

    _raw_len = 8
    _relevant_idx = 2

    button_id: ButtonId
    action: ButtonAction

    @classmethod
    def from_raw_bytes(cls, data: Union[bytearray, bytes]) -> "ButtonEvent":
        """Decomposes raw bytes into its components and returns them as a :class:`ButtonEvent`"""
        assert len(data) == cls._raw_len, f"Expected to get {cls._raw_len} bytes, got {len(data)}"
        button_id, action = unpack_from("<BB", data, 0)
        return cls(button_id=ButtonId(button_id), action=ButtonAction(action))


def flatten_nested_dataclass_fields(d: Union[Field, type], name: Optional[str]) -> list:
    """Analogous to `flatten_nested_dicts`, but returns the key names and works on the static class, not an instance

    Args:
        d: A dataclass e.g. `DataPackage`
        name: The prefix to this level, or more specifically, the key for the next higher level. If `None`, no prefix
              is used

    Returns:
        Returns the concatenated keys

    Examples:
        >>> @dataclass
        ... class A:
        ...     x: float
        ...     y: float
        >>> flatten_nested_dataclass_fields(A, None)
        ['x', 'y']
        >>> @dataclass
        ... class B:
        ...     a: A
        ...     b: int
        >>> flatten_nested_dataclass_fields(B, None)
        ['a_x', 'a_y', 'b']
    """
    results = []
    if hasattr(d, "__dataclass_fields__"):
        for k, v in d.__dataclass_fields__.items():
            curr_name = f"{name}_{k}" if name is not None else k
            curr_val = flatten_nested_dataclass_fields(v, curr_name)
            results.extend(curr_val)
    elif isinstance(d, Field):
        curr_val = flatten_nested_dataclass_fields(d.type, name)
        results.extend(curr_val)
    else:
        results.append(name)

    return results


def process_byte_data(raw_bytes: Union[bytearray, bytes]) -> Union[ButtonEvent, DataPackage, RawDataPackage]:
    """Factory function that takes raw bytes and translates into a button event or data

    Args:
        raw_bytes: The input raw bytes

    Returns:
        The resulting button event or data
    """
    q = PackageMetadata.from_raw_bytes(raw_bytes)
    raw_bytes_data = PackageMetadata.split_out_data_from_metadata(raw_bytes)

    if q.is_data():
        package = DataPackage.from_raw_bytes(raw_bytes_data)
    elif q.is_button():
        package = ButtonEvent.from_raw_bytes(raw_bytes_data)
    elif q.is_raw_gyro_accel():
        package = RawDataPackage.from_raw_bytes(raw_bytes_data)
    else:
        raise ValueError(f"Unknown value for q.id={q.id}")

    return package



import math
from dataclasses import dataclass


@dataclass(frozen=True)
class Point3d:
    x: float
    y: float
    z: float

    def __sub__(self, other):
        return Point3d(x=self.x - other.x, y=self.y - other.y, z=self.z - other.z)

    def as_dict(self, prefix: str = ""):
        return {f"{prefix}x": self.x, f"{prefix}y": self.y, f"{prefix}z": self.z}


@dataclass(frozen=True)
class Euler3d:
    roll: float
    pitch: float
    yaw: float

    def as_dict(self, prefix: str = ""):
        return {f"{prefix}roll": self.roll, f"{prefix}pitch": self.pitch, f"{prefix}yaw": self.yaw}


@dataclass(frozen=True)
class Quaternion:
    w: float
    x: float
    y: float
    z: float

    @classmethod
    def from_point3d(cls, p: Point3d) -> "Quaternion":
        return cls(0, p.x, p.y, p.z)

    def to_point3d(self):
        return Point3d(self.x, self.y, self.z)

    def __mul__(self, other):
        w1, x1, y1, z1 = self.w, self.x, self.y, self.z
        w2, x2, y2, z2 = other.w, other.x, other.y, other.z

        w = w1 * w2 - x1 * x2 - y1 * y2 - z1 * z2
        x = w1 * x2 + x1 * w2 + y1 * z2 - z1 * y2
        y = w1 * y2 + y1 * w2 + z1 * x2 - x1 * z2
        z = w1 * z2 + z1 * w2 + x1 * y2 - y1 * x2
        return Quaternion(w, x, y, z)

    def conjugate(self):
        return Quaternion(self.w, -self.x, -self.y, -self.z)

    def normalize(self):
        norm = math.sqrt(sum([el**2 for el in [self.w, self.x, self.y, self.z]]))
        return Quaternion(self.w / norm, self.x / norm, self.y / norm, self.z / norm)

    def as_dict(self, prefix: str = ""):
        return {f"{prefix}w": self.w, f"{prefix}x": self.x, f"{prefix}y": self.y, f"{prefix}z": self.z}


def rotate_vector(p: Point3d, q: Quaternion) -> Point3d:
    """Rotate point p by quaternion q"""
    q = q.normalize()
    p = Quaternion.from_point3d(p)
    p_rot = q * p * q.conjugate()
    return p_rot.to_point3d()
